
/*
 * This file is part of EKG/gEKG, a software package to record EKG.
 * Copyright  2003 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "EKGsource.h"



// ----------------------------------------------------------------------
//
// Konstruktor und Destruktor
//
// ----------------------------------------------------------------------



EKGsource::~EKGsource()
{
  if ( close( SerialPort ) != 0 )
    throw string( "EKGstream::~EKGstream : can't close serial port.\n" );
}



// ----------------------------------------------------------------------
//
// public Methoden
//
// ----------------------------------------------------------------------



/*
 * ffnet serielle Schnittstelle 
 */
void EKGsource::serialOpen ( const char * pathname )
{
  // Seriellen Port ffnen
  SerialPort = open( pathname, O_RDWR | O_NOCTTY );
  if ( SerialPort == -1 )
    throw string( "EKGsource::serialOpen : can't open device " ) + pathname;
  // Terminal auf Rohdaten schalten
  if ( tcgetattr( SerialPort, &SerialIOs ) != 0 )
    throw string( "EKGsource::serialOpen : can't get attributes of" ) + 
          pathname + ".";
  SerialIOs.c_iflag = IGNBRK | IGNPAR;
  SerialIOs.c_oflag = 0;
  SerialIOs.c_cflag = CS8 | CREAD | HUPCL | CLOCAL;
  SerialIOs.c_lflag = 0;
  // Geschwindigkeit setzen
  int error = cfsetospeed( &SerialIOs,  B9600 );
  error += cfsetispeed( &SerialIOs,  B9600 ); // input speed = output speed
  error += tcsetattr( SerialPort, TCSANOW, &SerialIOs );
  if ( error != 0 )
    throw string( "EKGsource::serialOpen : can't set attributes of " ) + 
          pathname + ".";
}


/*
 * startet Datenquelle
 */
void EKGsource::start()
{
  if ( sending ) return;
  if ( nextDevice == NULL ) return;
  sending = true;
  pthread_create( &senderID, NULL, &EKGsource::Sender, this );
}

/*
 * Sender anhalten
 */
void EKGsource::stop()
{
  if ( !sending ) return;
  sending = false;
  pthread_cancel( senderID );
  pthread_join( senderID, NULL );
}

/*
 * Zeitbegrenztes senden
 */
void EKGsource::timer( int t )
{
  if ( sending ) return;
  if ( nextDevice == NULL ) return;
  sending = true;
  flush();
  synchronise();
  time_t start = time( NULL );
  time_t stop = start + t;
  while( time( NULL ) <= stop )
  {
    fillBasket();
    nextDevice->input( dataBasket );
  }
  sending = false;
}




/*
 * Gibt Rohdaten auf stdout aus.
 */
void EKGsource::raw()
{
  while ( true )
  {
    for ( int i = 0; i < 4; i++ )
      cout << readByte() << ' ';
    cout << endl;
  }
}



/*
 * Empfnger anschlieen
 */
void EKGsource::connectTo( EKGdevice * d )
{
  if ( sending )
   throw string("EKGsource::connectTo : Sender's on, reconnection a bad idea.");
 nextDevice = d;
}



// ----------------------------------------------------------------------
//
// private Methoden
//
// ----------------------------------------------------------------------

// ----------------------------------------------------------------------
// Serielle Schnittstelle
// ----------------------------------------------------------------------

/*
 * Liest genau ein Byte vom seriellen Port. 
 */
int EKGsource::readByte()
{
  // Was ist, wenn keine Daten da sind (Hardware aus)?
  // Wartet read? Ist returnvalue = 0? Warum wird _manchmal_ 0 als Wert
  // zurckgegeben?
  unsigned char byte;
  read( SerialPort, & byte, 1 );
  return static_cast<int>(byte);
}

// ----------------------------------------------------------------------
// Datenstrom
// ----------------------------------------------------------------------

/*
 * Der Datenstrom enthlt drei Mewerte, gefolgt von einem groen
 * konstanten Wert.  Diese Methode synchronisiert den Datenstrom so, da
 * der grte Wert immer am Ende eines Wertequartetts steht.
 */
void EKGsource::synchronise()
{
  int values[4] = { 0, 0, 0, 0 };
  int maxVal = 0;
  
  // 4 Byte wegschmeien, denn beim Einschalten kann Mist gesendet werden.
  for ( int i = 0; i < 4; i++ ) readByte();
  // Werte Sammeln
  for ( int k = 0; k < 100; k++ )
    for ( int i = 0; i < 4; i++ ) 
      values[i] += readByte();
  // grten Wert suchen
  for ( int i = 1; i < 4; i++ )
    if ( values[i] > values[ maxVal ] )
      maxVal = i;
  // Synchronisieren: Werte lesen, bis der grte gelesen wurde.
  for ( int i = 0; i < 4; i++ )
  {
    readByte();
    if ( maxVal == i ) break;
  }
}


/*
 * Verwirft alle empfangenen aber noch nicht gelesenen Daten.
 */
void EKGsource::flush()
{
  tcflush( SerialPort, TCIFLUSH );
}


/*
 * ,,Datenkrbchen'' (Datentriplett) zum Senden bereitstellen.
 */
void EKGsource::fillBasket()
{
  for ( int i = 0; i < 3; i++ )
    dataBasket[i] = EKG_amp[i] * ( readByte() - EKG_zero[i] );
  // synchronisations-Wert lesen
  readByte();
}


// ----------------------------------------------------------------------
// Sender
// ----------------------------------------------------------------------

/*
 * Sender im Thread
 */
void * EKGsource::Sender( void * Me )
{
  EKGsource * me = static_cast<EKGsource*>(Me);
  pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, NULL );
  pthread_setcanceltype( PTHREAD_CANCEL_DEFERRED, NULL );
  me->flush();
  me->synchronise();
  while ( true )
    try
    {
      me->fillBasket();
      me->nextDevice->input( me->dataBasket );
    }
    catch (string s) 
    { 
      cerr << "Sender: " << s << endl; 
    }
  return NULL;
}


